////////////////////////////////////////////////////////////////////////////
/*

	Susie Plug-in for PMS(*.pms) (C)AliceSoft

*/
////////////////////////////////////////////////////////////////////////////
#ifndef IF_PMS_H__
#define IF_PMS_H__

#include "SpiBase.h"




//////////////////////////////////////////////////////////////////
//	PMS vOC
class SpiPMS : public SpiBase
{
#pragma	pack(1)

	//////////////////////////////////////////////////////////////////
	// PMSwb_
	struct PMSInfo
	{
		ULONG id;	// "PM", 0x0001
		USHORT sizeOfThis1;	// 0x0030
		BYTE bitCount;	// 0x10
		BYTE bitCount2;	// 0x10
		ULONG reserved1[4];
		long width;
		long height;
		ULONG sizeOfThis2;	// 0x00000030
		DWORD reserved2[3];
	};
#pragma pack()

	static const ULONG m_id	= 0x00014D50;

public:
	// Ήt@CmF
	static int IsSupported(BYTE* data)
	{
		PMSInfo* pInfo	= reinterpret_cast<PMSInfo*>(data);
		if (pInfo->id == m_id && pInfo->sizeOfThis2 == 0x30 && pInfo->bitCount == 16)
		{
			return 1;
		}
		return 0;
	}

	// 摜t@CɊւ𓾂
	int GetPictureInfo(BYTE* buf, DWORD len, PictureInfo *lpInfo)
	{
		const PMSInfo& info	= *reinterpret_cast<PMSInfo*>(buf);
		if (len < sizeof(PMSInfo))
		{
			return SPI_NOT_SUPPORT;
		}

		lpInfo->width		= info.width;
		lpInfo->height		= info.height;
		lpInfo->left		= 0;
		lpInfo->top			= 0;
		lpInfo->x_density	= 0;
		lpInfo->y_density	= 0;
		lpInfo->hInfo		= NULL;
		lpInfo->colorDepth	= (info.bitCount == 16) ? 24 : info.bitCount;

		return SPI_ALL_RIGHT;
	}

	// 摜WJ
	int GetPicture(BYTE* buf, DWORD len, HANDLE *pHBInfo, HANDLE *pHBm, SPI_CALLBACK spiCallBack, long lData)
	{
		// 摜̎擾
		const PMSInfo& info	= *reinterpret_cast<PMSInfo*>(buf);
		if (len < sizeof(PMSInfo))
		{
			return SPI_NOT_SUPPORT;
		}

		if (info.bitCount != 16)
		{
#ifdef	_DEBUG
			MsgBoxF("m̃f[^", "bitCount = %d", info.bitCount);
#endif
			return SPI_NOT_SUPPORT;
		}

		// BMP̃Zbg
		DWORD bmpInfoSize	= sizeof(BITMAPINFOHEADER);
		LocalHeap lhInfo;
		BITMAPINFO* pInfo	= static_cast<BITMAPINFO*>(lhInfo.Alloc(bmpInfoSize));
		if (pInfo == NULL)
			return SPI_NO_MEMORY;

		GetBitmapInfo(*pInfo, info);

		DWORD rowBytes	= ((info.width * pInfo->bmiHeader.biBitCount >> 3) + 3) & ~3;
		DWORD imageSize	= rowBytes * info.height;	// BMP4oCgƂɐ񂳂B

		// C[W̓WJ
		LocalHeap lhImage, lhBuffer;
		BYTE* pImage	= static_cast<BYTE*>(lhImage.Alloc(imageSize));
		BYTE* pBuffer	= static_cast<BYTE*>(lhBuffer.Alloc(imageSize));
		if (pImage == NULL || pBuffer == NULL)
			return SPI_NO_MEMORY;

		ExtractImage(pBuffer, &buf[sizeof(PMSInfo)], info);
		DIBInvert(pImage, pBuffer, pInfo->bmiHeader);

		// AbN
		*pHBInfo	= lhInfo.Detach();	// SusieɕԂnhZbg
		*pHBm		= lhImage.Detach();
		return SPI_ALL_RIGHT;
	}

private:
	// PMSwb_ - >BMPwb_
	void GetBitmapInfo(BITMAPINFO& bmi, const PMSInfo& info)
	{
		bmi.bmiHeader.biSize			= sizeof(BITMAPINFOHEADER);
		bmi.bmiHeader.biWidth			= info.width;
		bmi.bmiHeader.biHeight			= info.height;
		bmi.bmiHeader.biPlanes			= 1;
		bmi.bmiHeader.biBitCount		= (info.bitCount == 16) ? 24 :info.bitCount;
//		bmi.bmiHeader.biCompression		= BI_RGB;
//		bmi.bmiHeader.biSizeImage		= 0;
//		bmi.bmiHeader.biXPelsPerMeter	= 0;
//		bmi.bmiHeader.biYPelsPerMeter	= 0;
//		bmi.bmiHeader.biClrUsed			= 0;
//		bmi.bmiHeader.biClrImportant	= 0;
	}

	// 摜̓WJ
	void ExtractImage(BYTE* dest, const BYTE* src, const PMSInfo& info)
	{
		long pixelCount	= info.width * info.height;
		int read;
		while (pixelCount)
		{
			int format	= *src++ - 0xF8;
			switch (format)
			{
			case 0:
				// 0x1786
				pixelCount--;
				Decode1(dest, src);
				dest	= &dest[3];
				src		= &src[2];
				break;
			case 1:
				// 0x179C
			{
				read	= *src++ + 1;
				pixelCount	-= read;
				BYTE data[2];
				data[1]	= *src++;
				for (int i = 0;i < read;i++)
				{
					data[0]	= *src++;
					Decode2(dest, data);
					dest	= &dest[3];
				}
			}
				break;
			case 2:
				// 0x17E4
				pixelCount--;
				dest[0]	= dest[-info.width*3 + 3];
				dest[1]	= dest[-info.width*3 + 4];
				dest[2]	= dest[-info.width*3 + 5];
				dest	= &dest[3];
				break;
			case 3:
				// 0x1809
				pixelCount--;
				dest[0]	= dest[-info.width*3 - 3];
				dest[1]	= dest[-info.width*3 - 2];
				dest[2]	= dest[-info.width*3 - 1];
				dest	= &dest[3];
				break;
			case 4:
				// 0x182C
				read	= *src++ + 2;
				pixelCount	-= read * 2;
				for (int i = 0;i < read;i++)
				{
					Decode1(dest, src);
					Decode1(&dest[3], &src[2]);
					dest	= &dest[6];
				}
				src		= &src[4];
				break;
			case 5:
				// 0x1874
				read	= *src++ + 3;
				pixelCount	-= read;
				for (int i = 0;i < read;i++)
				{
					Decode1(dest, src);
					dest	= &dest[3];
				}
				src		= &src[2];
				break;
			case 6:
				// 0x18AB
				read	= *src++ + 2;
				pixelCount	-= read;
				for (int i = 0;i < read;i++)
				{
					dest[0]	= dest[-info.width*3*2];
					dest[1]	= dest[-info.width*3*2 + 1];
					dest[2]	= dest[-info.width*3*2 + 2];
					dest	= &dest[3];
				}
				break;
			case 7:
				// 18EC
				read	= *src++ + 2;
				pixelCount	-= read;
				for (int i = 0;i < read;i++)
				{
					dest[0]	= dest[-info.width*3];
					dest[1]	= dest[-info.width*3 + 1];
					dest[2]	= dest[-info.width*3 + 2];
					dest	= &dest[3];
				}
				break;
			default:
				// goto 0x192B
				pixelCount--;
				Decode1(dest, --src);
				dest	= &dest[3];
				src		= &src[2];
				break;
			}
		}
	}

	// sNZ̓WJP
	void Decode1(BYTE* dest, const BYTE* src)
	{
		USHORT ax	= *reinterpret_cast<const USHORT*>(src);
		BYTE a	= ax & 0x1F;
		dest[0]	= (a >> 2) + (a << 3);
		BYTE b	= (ax >> 6) & 0x1F;
		dest[1]	= (b >> 3) + (b << 3) + (((BYTE)ax >> 3) & 0x04);
		BYTE c	= ax >> 11;
		dest[2]	= (c >> 2) + (c << 3);
	}

	// sNZ̓WJQ
	void Decode2(BYTE* dest, const BYTE* src)
	{
		USHORT ax	= *reinterpret_cast<const USHORT*>(src);
		BYTE a	= (ax & 0x03) + ((ax >> 6) & 0x1C);
		dest[0]	= (a >> 2) + (a << 3);
		BYTE b	= ((ax >> 8) & 0x18) + ((ax >> 3) & 0x07);
		dest[1]	= (b >> 3) + (b << 3) + ((BYTE)ax & 0x04);
		BYTE c	= ((ax >> 6) & 0x03) + ((ax >> 11) & 0x1C);
		dest[2]	= (c >> 2) + (c << 3);
	}
};

#endif	// IF_PMS_H__