////////////////////////////////////////////////////////////////////////////
/*

	Susie Plug-in for QNT(*.qnt) (C)AliceSoft

*/
////////////////////////////////////////////////////////////////////////////
#ifndef IF_QNT_H__
#define IF_QNT_H__

#include "SpiBase.h"
#include "zlib.h"

//////////////////////////////////////////////////////////////////
//	QNTwb_
#pragma	pack(1)
// DWORD id;	// "QNT"
// DWORD ver;	// 0-2 mF
// DWORD headerSize	// 0x44 version 0ɂ݂͑Ȃ
// QNTInfo info;

struct QNTInfo
{
	DWORD reserved1;
	DWORD reserved2;
	long width;
	long height;
	DWORD bitCount;
	DWORD reserved3;
	DWORD imageSize;
	DWORD alphaSize;
};
#pragma pack()


//////////////////////////////////////////////////////////////////
//	QNT vOC
class SpiQNT : public SpiBase
{
public:
	// 錾
	enum {
		ID_NOT_SUPPORT,
		ID_QNT		= 0x00544E51,	// ID
		QNT_MARGIN	= 4096
	};

	// Ήt@CmF
	static int IsSupported(BYTE* data)
	{
		DWORD* pID	= reinterpret_cast<DWORD*>(data);
		if (pID[0] == ID_QNT)
			return ID_QNT;
		return ID_NOT_SUPPORT;
	}

	// 摜t@CɊւ𓾂
	int GetPictureInfo(BYTE* buf, DWORD len, PictureInfo *lpInfo)
	{
		DWORD size;
		const QNTInfo& info	= GetQNTHeader(buf, size);

		lpInfo->width		= info.width;
		lpInfo->height		= info.height;
		lpInfo->left		= 0;
		lpInfo->top			= 0;
		lpInfo->x_density	= 0;
		lpInfo->y_density	= 0;
		lpInfo->hInfo		= NULL;
		lpInfo->colorDepth	= static_cast<short>(info.bitCount);

		if (info.alphaSize && !_config.m_bAlphaBrend)
			lpInfo->colorDepth	= 0x20;

		return SPI_ALL_RIGHT;
	}

	// 摜WJ
	int GetPicture(BYTE* buf, DWORD len, HANDLE *pHBInfo, HANDLE *pHBm, SPI_CALLBACK spiCallBack, long lData)
	{
		// 摜̎擾
		DWORD qntHeaderSize;
		QNTInfo qntInfo	= GetQNTHeader(buf, qntHeaderSize);

		if (qntInfo.bitCount != 0x18 || qntInfo.imageSize == 0)
		{
#ifdef	_DEBUG
			// ςC++̂ςĂ{GA\cg05002.QNT: alphaf[^܂܂Ȃςȃt@C
			MsgBoxF("m̃f[^", "bitCount = %d, imageSize = %d", qntInfo.bitCount, qntInfo.imageSize);
#endif
			return SPI_NOT_SUPPORT;
		}

		// BMP̃Zbg
		const DWORD bmpInfoSize	= sizeof(BITMAPINFOHEADER);
		LocalHeap lhInfo;
		BITMAPINFO* pInfo	= reinterpret_cast<BITMAPINFO*>(lhInfo.Alloc(bmpInfoSize));
		if (pInfo == NULL)
			return SPI_NO_MEMORY;

		GetBitmapInfo(*pInfo, qntInfo);

		DWORD rowBytes	= ((qntInfo.width * pInfo->bmiHeader.biBitCount >> 3) + 3) & ~3;
		DWORD imageSize	= rowBytes * qntInfo.height;	// BMP4oCgƂɐ񂳂B

		// C[W̓WJ
		LocalHeap lhImage;
		BYTE* pImage	= reinterpret_cast<BYTE*>(lhImage.Alloc(imageSize));
		if (pImage == NULL)
			return SPI_NO_MEMORY;

		bool bOk	= false;
		if (pInfo->bmiHeader.biBitCount == 0x20)
		{
			bOk	= GetDIBImage32(pImage, qntInfo, &buf[qntHeaderSize]);
		} else { 
			bOk	= GetDIBImage24(pImage, qntInfo, &buf[qntHeaderSize]);
		}
		if (bOk)
		{
			// AbN or 
			*pHBInfo	= lhInfo.Detach();	// SusieɕԂnhZbg
			*pHBm		= lhImage.Detach();
			return SPI_ALL_RIGHT;
		}
		return SPI_NOT_SUPPORT;
	}

private:
	// QNTwb_ւ̎QƂԂ
	const QNTInfo& GetQNTHeader(BYTE* buffer, DWORD& headerSize)
	{
		DWORD version	= *reinterpret_cast<DWORD*>(&buffer[4]);
		if (version == 0)
		{
			headerSize	= 48;
			return *reinterpret_cast<QNTInfo*>(&buffer[8]);
		}
//		else	// version == 2܂ŁB͕s
		{
			headerSize	= *reinterpret_cast<DWORD*>(&buffer[8]);
			return *reinterpret_cast<QNTInfo*>(&buffer[12]);
		}
	}

	// QNTwb_ - >BMPwb_
	void GetBitmapInfo(BITMAPINFO& bmi, const QNTInfo& qntInfo)
	{
		WORD bitCount	= 0x18;
		if (qntInfo.alphaSize && !_config.m_bAlphaBrend)
			bitCount	= 0x20;

		bmi.bmiHeader.biSize			= sizeof(BITMAPINFOHEADER);
		bmi.bmiHeader.biWidth			= qntInfo.width;
		bmi.bmiHeader.biHeight			= qntInfo.height;
		bmi.bmiHeader.biPlanes			= 1;
		bmi.bmiHeader.biBitCount		= bitCount;
//		bmi.bmiHeader.biCompression		= BI_RGB;
//		bmi.bmiHeader.biSizeImage		= 0;
//		bmi.bmiHeader.biXPelsPerMeter	= 0;
//		bmi.bmiHeader.biYPelsPerMeter	= 0;
//		bmi.bmiHeader.biClrUsed			= 0;
//		bmi.bmiHeader.biClrImportant	= 0;
	}

	// `l̓WJ
	bool ExtractAlpha(BYTE* alpha, DWORD alphaSize, BYTE* src, const QNTInfo& qntInfo)
	{
		if (Z_OK != uncompress(alpha, &alphaSize, src, qntInfo.alphaSize))
			return false;

		// ƍ̃sNZƂ̐F
		long width	= (qntInfo.width + 1) & ~1;	// 2oCg
		for (long j = 0;j < width - 1;j++)
		{
			// 1sڂ͍Ƃ̍̂
			alpha[j + 1]	= alpha[j] - alpha[j + 1];
		}

		for (long i = 0;i < qntInfo.height - 1;i++)
		{
			// 2sڈȍ~
			BYTE* prev	= alpha;
			alpha	= &alpha[width];
			alpha[0]	= prev[0] - alpha[0];

			for (long j = 0;j < width - 1;j++)
			{
				alpha[j + 1]	= (alpha[j] + prev[j + 1]) / 2 - alpha[j + 1];
			}
		}
		return true;
	}

	// C[WDIBɓWJ
	bool GetDIBImage24(BYTE* pImage, const QNTInfo& qntInfo, BYTE* src)
	{
		DWORD rowBytes	= (qntInfo.width * 3 + 3) & ~3;
		DWORD destSize	= rowBytes * qntInfo.height + QNT_MARGIN;	// ܂AmȃTCY킩
		LocalHeap lhBuffer, lhAlpha;
		BYTE* pBuffer	= reinterpret_cast<BYTE*>(lhBuffer.Alloc(destSize));
		BYTE* pAlpha	= NULL;
		if (qntInfo.alphaSize)
		{
			// `l̓WJ
			DWORD alphaSize	= (qntInfo.width + 1) * qntInfo.height + QNT_MARGIN;
			pAlpha	= reinterpret_cast<BYTE*>(lhAlpha.Alloc(alphaSize));
			if (pBuffer == NULL || pAlpha == NULL)
				return false;

			if (!ExtractAlpha(pAlpha, alphaSize, &src[qntInfo.imageSize], qntInfo))
				return false;
		} else {
			if (pBuffer == NULL)
				return false;
		}
		// ZLIB.uncompress()ŃC[WWJ
		if (Z_OK != uncompress(pBuffer, &destSize, src, qntInfo.imageSize))
			return false;

		// BGR, BGR, ... ̃I[_[ɕёւ
		pImage	= &pImage[rowBytes * (qntInfo.height - 1)];	// ㉺t]
		for (int color = 0;color < 3;color++)
		{
			BYTE* pDest	= &pImage[color];
			for (long i = 0;i < qntInfo.height - 1;i += 2)
			{
				BYTE* pDest1	= pDest;
				BYTE* pDest2	= pDest - rowBytes;
				for (long j = 0;j < qntInfo.width - 1;j += 2)
				{
					pDest1[0]	= *pBuffer++;
					pDest2[0]	= *pBuffer++;
					pDest1[3]	= *pBuffer++;
					pDest2[3]	= *pBuffer++;
					pDest1	= &pDest1[6];
					pDest2	= &pDest2[6];
				}
				if (qntInfo.width & 1)	// ƖɃ_~[Ă
				{
					pDest1[0]	= pBuffer[0];
					pDest2[0]	= pBuffer[1];
					pBuffer	= &pBuffer[4];
				}
				pDest	-= rowBytes * 2;
			}
			if (qntInfo.height & 1)		// ƖɃ_~[sĂ
			{
				for (long j = 0;j < qntInfo.width - 1;j += 2)
				{
					pDest[0]	= pBuffer[0];
					pDest[3]	= pBuffer[2];
					pDest	= &pDest[6];
					pBuffer	= &pBuffer[4];
				}
				if (qntInfo.width & 1)
				{
					pDest[0]	= pBuffer[0];
					pBuffer	= &pBuffer[4];
				}
			}
		}
		// ƍ̃sNZƂ̐F
		BYTE* pDest	= pImage;
		for (long j = 0;j < qntInfo.width - 1;j++)
		{
			// 1sڂ͍Ƃ̍̂
			pDest[3]	= pDest[0] - pDest[3];
			pDest[4]	= pDest[1] - pDest[4];
			pDest[5]	= pDest[2] - pDest[5];
			pDest	= &pDest[3];
		}

		pDest	= pImage;
		for (long i = 0;i < qntInfo.height - 1;i++)
		{
			// 2sڈȍ~
			BYTE* pDest1	= pDest;
			pDest	-= rowBytes;
			BYTE* pDest2	= pDest;
			pDest2[0]	= pDest1[0] - pDest2[0];
			pDest2[1]	= pDest1[1] - pDest2[1];
			pDest2[2]	= pDest1[2] - pDest2[2];

			for (long j = 0;j < qntInfo.width - 1;j++)
			{
				pDest2[3]	= (pDest1[3] + pDest2[0]) / 2 - pDest2[3];
				pDest2[4]	= (pDest1[4] + pDest2[1]) / 2 - pDest2[4];
				pDest2[5]	= (pDest1[5] + pDest2[2]) / 2 - pDest2[5];
				pDest1	= &pDest1[3];
				pDest2	= &pDest2[3];
			}
		}
		// uh
		if (qntInfo.alphaSize)
		{
			for (long i = 0;i < qntInfo.height;i++)
			{
				pDest	= pImage;
				for (long j = 0;j < qntInfo.width;j++)
				{
					RGBQUAD& bgColor	= _config.m_rgb;
					if (pAlpha[j] == 0xFF)
					{}
					else if (pAlpha[j] == 0x00)
					{
						pDest[0]	= bgColor.rgbBlue;
						pDest[1]	= bgColor.rgbGreen;
						pDest[2]	= bgColor.rgbRed;
					} else {
						pDest[0]	= AlphaBlend(pDest[0], bgColor.rgbBlue, pAlpha[j]);
						pDest[1]	= AlphaBlend(pDest[1], bgColor.rgbGreen, pAlpha[j]);
						pDest[2]	= AlphaBlend(pDest[2], bgColor.rgbRed, pAlpha[j]);
					}
					pDest	= &pDest[3];
				}
				pImage	-= rowBytes;
				pAlpha	= &pAlpha[(qntInfo.width + 1) & ~1];	// 2oCg
			}			
		}
		return true;
	}

	// C[W32bit DIBɓWJ
	bool GetDIBImage32(BYTE* pImage, const QNTInfo& qntInfo, BYTE* src)
	{
		DWORD rowBytes	= qntInfo.width << 2;
		DWORD destSize	= (qntInfo.width + 1) * (qntInfo.height + 1) * 3 + QNT_MARGIN;	// Kɑ߂
		LocalHeap lhBuffer;
		BYTE* pBuffer	= reinterpret_cast<BYTE*>(lhBuffer.Alloc(destSize));
		if (pBuffer == NULL)
			return false;

		// ZLIB.uncompress()ŃC[WWJ
		if (Z_OK != uncompress(pBuffer, &destSize, src, qntInfo.imageSize))
			return false;

		// BGR, BGR, ... ̃I[_[ɕёւ
		pImage	= &pImage[rowBytes * (qntInfo.height - 1)];	// ㉺t]
		BYTE* pSrc	= pBuffer;
		for (int color = 0;color < 3;color++)
		{
			BYTE* pDest	= &pImage[color];
			for (long i = 0;i < qntInfo.height - 1;i += 2)
			{
				BYTE* pDest1	= pDest;
				BYTE* pDest2	= pDest - rowBytes;
				for (long j = 0;j < qntInfo.width - 1;j += 2)
				{
					pDest1[0]	= *pSrc++;
					pDest2[0]	= *pSrc++;
					pDest1[4]	= *pSrc++;
					pDest2[4]	= *pSrc++;
					pDest1	= &pDest1[8];
					pDest2	= &pDest2[8];
				}
				if (qntInfo.width & 1)	// ƖɃ_~[Ă
				{
					pDest1[0]	= pSrc[0];
					pDest2[0]	= pSrc[1];
					pSrc	= &pSrc[4];
				}
				pDest	-= rowBytes * 2;
			}
			if (qntInfo.height & 1)		// ƖɃ_~[sĂ
			{
				for (long j = 0;j < qntInfo.width - 1;j += 2)
				{
					pDest[0]	= pSrc[0];
					pDest[4]	= pSrc[2];
					pDest	= &pDest[8];
					pSrc	= &pSrc[4];
				}
				if (qntInfo.width & 1)
				{
					pDest[0]	= pSrc[0];
					pSrc	= &pSrc[4];
				}
			}
		}

		// ƍ̃sNZƂ̐F
		BYTE* pDest	= pImage;
		for (long j = 0;j < qntInfo.width - 1;j++)
		{
			// 1sڂ͍Ƃ̍̂
			pDest[4]	= pDest[0] - pDest[4];
			pDest[5]	= pDest[1] - pDest[5];
			pDest[6]	= pDest[2] - pDest[6];
			pDest	= &pDest[4];
		}

		pDest	= pImage;
		for (long i = 0;i < qntInfo.height - 1;i++)
		{
			// 2sڈȍ~
			BYTE* pDest1	= pDest;
			pDest	-= rowBytes;
			BYTE* pDest2	= pDest;
			pDest2[0]	= pDest1[0] - pDest2[0];
			pDest2[1]	= pDest1[1] - pDest2[1];
			pDest2[2]	= pDest1[2] - pDest2[2];

			for (long j = 0;j < qntInfo.width - 1;j++)
			{
				pDest2[4]	= (pDest1[4] + pDest2[0]) / 2 - pDest2[4];
				pDest2[5]	= (pDest1[5] + pDest2[1]) / 2 - pDest2[5];
				pDest2[6]	= (pDest1[6] + pDest2[2]) / 2 - pDest2[6];
				pDest1	= &pDest1[4];
				pDest2	= &pDest2[4];
			}
		}

		// `l̃Zbg
		// ZLIB.uncompress()Ń`lWJ
		if (Z_OK != uncompress(pBuffer, &destSize, &src[qntInfo.imageSize], qntInfo.alphaSize))
			return false;

		// ƍ̃sNZƂ̐F
		pDest	= &pImage[3];
		pDest[0]	= *pBuffer++;
		for (long j = 0;j < qntInfo.width - 1;j++)
		{
			// 1sڂ͍Ƃ̍̂
			pDest[4]	= pDest[0] - *pBuffer++;
			pDest	= &pDest[4];
		}
		if (qntInfo.width & 1)	// 2oCgŐA̂Ƃ1oCg̃_~[
			pBuffer++;
		pImage	= &pImage[3];
		for (long i = 0;i < qntInfo.height - 1;i++)
		{
			// 2sڈȍ~
			BYTE* prev	= pImage;
			pImage	-= rowBytes;
			pDest	= pImage;
			pDest[0]	= prev[0] - *pBuffer++;

			for (long j = 0;j < qntInfo.width - 1;j++)
			{
				pDest[4]	= (pDest[0] + prev[4]) / 2 - *pBuffer++;
				pDest	= &pDest[4];
				prev	= &prev[4];
			}
			if (qntInfo.width & 1)
				pBuffer++;
		}
		return true;
	}
};


#endif // IF_QNT_H__
